from dataclasses import dataclass, field
import numpy as np


@dataclass
class IntersectionResult:
    """
    A class to store the intersection results of a ray with a mesh.

    Attributes
    ----------
    coords : np.ndarray
        The coordinates of the intersection points.
    tri_ids : np.ndarray
        The triangle ids of the intersected triangles.
    distances : np.ndarray
        The distances from the ray origin to the intersection points.
    reflections : np.ndarray
        The reflection vectors at the intersection points.
    """

    coords: np.ndarray = field(default_factory=lambda: np.empty((0, 3)))
    tri_ids: np.ndarray = field(default_factory=lambda: np.empty((0,)))
    distances: np.ndarray = field(default_factory=lambda: np.empty((0,)))
    reflections: np.ndarray = field(default_factory=lambda: np.empty((0, 3)))

    def __len__(self) -> int:
        return len(self.coords)

    @property
    def hit_mask(self) -> np.ndarray:
        """
        Returns a boolean mask of hits
        """
        return ~np.isnan(self.coords[:, 0])

    @property
    def num_hits(self) -> int:
        """
        Returns the number of intersection points.
        :return: int
        """
        return (~np.isnan(self.coords[:, 0])).sum()



import pytest

from pyraymesh import Mesh
import numpy as np

def mesh_plane():
    vertices = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]])
    faces = np.array([[0, 1, 2], [2, 3, 0]])
    return Mesh(vertices, faces)

def test_simple_reflecton():
    m = mesh_plane()
    m.build("medium")
    ray_origin = [[0.5, 0.5, 1], [0.5, 0.5, -1]]
    ray_direction = [[0, 0, -1], [0, 0, 1]]
    result = m.intersect(ray_origin, ray_direction, calculate_reflections=True)
    assert len(result) == 2
    assert len(result.reflections) == 2
    assert np.allclose(result.reflections[0], [0, 0, 1])
    assert np.allclose(result.reflections[1], [0, 0, -1])

def test_angled_reflection():
    m = mesh_plane()
    m.build("medium")
    ray_origin = [[0.5, 0.5, 1], [0.5, 0.5, -1]]
    ray_direction = [[-0.1,-0.1,-1], [-0.1,-0.1,1]]
    result = m.intersect(ray_origin, ray_direction, calculate_reflections=True)
    assert len(result) == 2
    assert len(result.reflections) == 2
    assert np.linalg.norm(result.reflections[0]) == 1
    assert np.linalg.norm(result.reflections[1]) == 1
    expected_reflection1 = np.array([-0.1, -0.1, 1])
    expected_reflection2 = np.array([-0.1, -0.1, -1])
    expected_reflection1 /= np.linalg.norm(expected_reflection1)
    expected_reflection2 /= np.linalg.norm(expected_reflection2)

    assert np.allclose(result.reflections[0], expected_reflection1)
    assert np.allclose(result.reflections[1], expected_reflection2)

def test_miss():
    m = mesh_plane()
    m.build()
    ray_origin = [[0.5, 0.5, 1]]
    ray_direction = [[0, 0, 1]]
    result = m.intersect(ray_origin, ray_direction, calculate_reflections=True)
    assert result.num_hits == 0
    assert len(result.reflections) == 1
    assert np.all(np.isnan(result.reflections[0]))


import numpy as np

from . import _bvh_bind_ext
from .IntersectionResult import IntersectionResult
from typing import List, Iterable, Union


def _prep_rays(ray_origin, ray_direction):
    ray_origin = np.array(ray_origin, dtype=np.float32)
    ray_direction = np.array(ray_direction, dtype=np.float32)
    if len(ray_origin.shape) == 1:
        if len(ray_origin) != 3:
            raise ValueError("ray_origin must have 3 elements")
        ray_origin = ray_origin[np.newaxis, :]
    if len(ray_direction.shape) == 1:
        if len(ray_direction) != 3:
            raise ValueError("ray_direction must have 3 elements")
        ray_direction = ray_direction[np.newaxis, :]
    if len(ray_origin) == 1 and len(ray_direction) > 1:
        ray_origin = np.repeat(ray_origin, len(ray_direction), axis=0)
    if len(ray_direction) == 1 and len(ray_origin) > 1:
        ray_direction = np.repeat(ray_direction, len(ray_origin), axis=0)
    if not len(ray_origin) == len(ray_direction):
        raise ValueError(
            "ray_origin and ray_direction must have the same length or one of them must have length 1"
        )
    return ray_origin, ray_direction


class Mesh:
    def __init__(
        self, vertices: Iterable[float], faces: Union[Iterable[int], None] = None
    ):
        """
        Initializes the Mesh object with vertices and optional faces.

        Parameters:
        vertices (array-like): An array of vertex coordinates.
        faces (array-like, optional): An array of face indices. Defaults to None. If None, the vertices are assumed to be a list of triangles.
        """
        self.vertices = vertices
        self.faces = faces
        self._normalize_mesh_data()
        self._bvh = None
        self.robust = True

    def _normalize_mesh_data(self):
        self.vertices = np.array(self.vertices, dtype=np.float32)
        if self.faces is not None:
            self.faces = np.array(self.faces, dtype=np.int32)
        if len(self.vertices.shape) == 1:
            self.vertices = self.vertices.reshape(-1, 3)
        if self.faces is not None and len(self.faces.shape) == 1:
            self.faces = self.faces.reshape(-1, 3)
        if self.faces is None:
            self.faces = np.arange(self.vertices.shape[0], dtype=np.int32).reshape(
                -1, 3
            )

    @property
    def is_built(self) -> bool:
        return self._bvh is not None

    def build(self, quality: str = "medium") -> None:
        """
        Builds the BVH (Bounding Volume Hierarchy) for the mesh with the specified quality.

        Parameters:
        quality (str): The quality level for building the BVH. Must be one of 'low', 'medium', or 'high'.
                       Defaults to 'medium'.

        Raises:
        ValueError: If the quality is not one of 'low', 'medium', or 'high'.
        """
        quality = quality.lower()
        if quality not in ["low", "medium", "high"]:
            raise ValueError("Quality must be one of 'low', 'medium' or 'high'")
        if len(self.vertices) == 0 or len(self.faces) == 0:
            raise ValueError("Mesh is empty")
        self._bvh = _bvh_bind_ext.build_bvh(self.vertices, self.faces, quality)

    def intersect(
        self,
        ray_origin: Iterable[Iterable[float]],
        ray_direction: Iterable[Iterable[float]],
        tnear: float = 0,
        tfar: float = np.finfo(np.float32).max,
        calculate_reflections: bool = False,
    ) -> IntersectionResult:
        """
        Intersects the rays with the mesh.

        Parameters:
        ray_origin (array-like): The origin points of the rays.
        ray_direction (array-like): The direction vectors of the rays.
        tnear (float, optional): The minimum distance along the ray to consider for intersections. Defaults to 0.
        tfar (float, optional): The maximum distance along the ray to consider for intersections. Defaults to np.inf.

        Returns:
        Hits: An object containing the intersection coordinates, triangle IDs, and distances.

        Raises:
        ValueError: If the BVH is not built and cannot be built with the specified quality.
        """
        if not self.is_built:
            print("BVH not built, building now with medium quality")
            self.build("medium")
            if not self.is_built:
                raise ValueError("failed to build BVH")
        ray_origin, ray_direction = _prep_rays(ray_origin, ray_direction)

        if calculate_reflections:
            coords, tri_ids, distances, reflections = _bvh_bind_ext.intersect_bvh(
                self._bvh, ray_origin, ray_direction, tnear, tfar, calculate_reflections, self.robust
            )
        else:
            coords, tri_ids, distances = _bvh_bind_ext.intersect_bvh(
                self._bvh, ray_origin, ray_direction, tnear, tfar, calculate_reflections, self.robust
            )
            reflections = np.empty((0, 3))

        return IntersectionResult(coords=coords, tri_ids=tri_ids, distances=distances, reflections=reflections)

    def occlusion(
        self,
        ray_origin: Iterable[float],
        ray_direction: Iterable[float],
        tnear=0,
        tfar=np.finfo(np.float32).max,
    ) -> np.ndarray:
        """
        Checks for occlusion along the rays with the BVH (Bounding Volume Hierarchy) of the mesh.

        Parameters:
        ray_origin (array-like): The origin points of the rays.
        ray_direction (array-like): The direction vectors of the rays.
        tnear (float, optional): The minimum distance along the ray to consider for occlusion. Defaults to 0.
        tfar (float, optional): The maximum distance along the ray to consider for occlusion. Defaults to np.inf.

        Returns:
        bool: (list of bool) A list of boolean values indicating whether the ray is occluded.

        Raises:
        ValueError: If the BVH is not built and cannot be built with the specified quality.
        """
        if not self.is_built:
            print("BVH not built, building now with medium quality")
            self.build("medium")
        ray_origin, ray_direction = _prep_rays(ray_origin, ray_direction)

        return np.array(
            _bvh_bind_ext.occlude_bvh(
                self._bvh, ray_origin, ray_direction, tnear, tfar, self.robust
            )
        )



import pytest

from pyraymesh import Mesh
import numpy as np


def mesh_plane():
    vertices = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]])
    faces = np.array([[0, 1, 2], [2, 3, 0]])
    return Mesh(vertices, faces)


def test_import_Mesh():
    m = Mesh([], [])
    assert m is not None
    assert isinstance(m, Mesh)

def test_empty_mesh():
    m = Mesh([], [])
    with pytest.raises(ValueError):
        m.build("medium")

def test_build_low_bvh():
    m = mesh_plane()
    m.build("low")
    assert m.is_built


def test_build_medium_bvh():
    m = mesh_plane()
    m.build("medium")
    assert m.is_built


def test_build_low_bvh():
    m = mesh_plane()
    m.build("high")
    assert m.is_built


def test_build_undef_bvh():
    m = mesh_plane()
    with pytest.raises(ValueError):
        m.build("undef")


def test_single_ray_intersection():
    m = mesh_plane()
    m.build("medium")
    ray_origin = [0.6, 0.7, 1]
    ray_direction = [0, 0, -1]
    result = m.intersect(ray_origin, ray_direction)

    assert len(result.coords) == 1
    assert result.coords[0][0] == 0.6
    assert result.coords[0][1] == 0.7
    assert result.coords[0][2] == 0

    assert len(result.tri_ids) == 1
    assert result.tri_ids[0] == 1

    assert len(result.distances) == 1
    assert result.distances[0] == 1

def test_origin_on_plane():
    m = mesh_plane()
    m.build("medium")
    ray_origin = [0.5, 0.5, 0]
    ray_direction = [0, 0, -1]
    result = m.intersect(ray_origin, ray_direction)
    assert len(result.coords) == 1
    assert result.coords[0][0] == 0.5
    assert result.coords[0][1] == 0.5
    assert result.coords[0][2] == 0

    result = m.intersect(ray_origin, ray_direction, tnear=0.01)
    assert len(result) == 1
    assert np.isnan(result.coords[0][0])



def test_tfar_intersection_miss():
    m = mesh_plane()
    m.build("medium")
    ray_origin = [0.5, 0.5, 1]
    ray_direction = [0, 0, -1]
    result = m.intersect(ray_origin, ray_direction, tfar=0.5)
    assert len(result) == 1
    assert np.isnan(result.coords[0][0])


def test_tnear_intersection_miss():
    m = mesh_plane()
    m.build("medium")
    ray_origin = [0.5, 0.5, 1]
    ray_direction = [0, 0, -1]
    result = m.intersect(ray_origin, ray_direction, tnear=2)
    assert len(result) == 1
    assert np.isnan(result.coords[0][0])

def test_angle_intersection():
    m = mesh_plane()
    m.build("medium")
    ray_origin = [0.5, 0.5, 1]
    ray_direction = [-0.1, -0.1, -1]
    result = m.intersect(ray_origin, ray_direction)
    assert len(result) == 1
    assert result.coords[0][0] == 0.4
    assert result.coords[0][1] == 0.4


def test_occulsion():
    m = mesh_plane()
    m.build("medium")
    ray_origin = [0.5, 0.5, 1]
    ray_direction = [0, 0, -1]
    occ = m.occlusion(ray_origin, ray_direction)
    assert len(occ) == 1
    assert occ[0] == True


def test_occulsion_miss():
    m = mesh_plane()
    m.build("medium")
    ray_origin = [0.5, 0.5, 1]
    ray_direction = [0, 0, 1]
    occ = m.occlusion(ray_origin, ray_direction)
    assert len(occ) == 1
    assert occ[0] == False

def test_hit_vertice():
    m = mesh_plane()
    m.build("medium")
    ray_origin = [0, 0, 0]
    ray_direction = [0, 0, 1]
    result = m.intersect(ray_origin, ray_direction)
    assert len(result) == 1
    assert result.coords[0][0] == 0


import pytest

from pyraymesh import Mesh
import numpy as np


def mesh_plane():
    vertices = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0]])
    faces = np.array([[0, 1, 2], [2, 3, 0]])
    return Mesh(vertices, faces)


def rays():
    ray_origin = np.array(
        [
            [0.01, 0.002, 1],
            [0.3, 0.4, 1],
            [0.5, 0.6, 1],
            [0.1, 0.2, 1],
            [0.3, 0.4, 1],
            [0.5, 0.6, 1],
        ]
    )
    ray_direction = np.array(
        [[0, 0, -1], [0, 0, -1], [0, 0, -1], [0, 0, -1], [0, 0, 1], [0, 0, 1]]
    )
    return ray_origin, ray_direction


def test_multiple_ray_intersect():
    m = mesh_plane()
    m.build("medium")
    ray_origin, ray_direction = rays()
    result = m.intersect(ray_origin, ray_direction)

    assert len(result) == 6
    assert (result.num_hits) == 4

    assert result.coords[0][0] == 0.01
    assert result.coords[0][1] == 0.002
    assert result.coords[0][2] == 0
    assert result.coords[1][0] == 0.3
    assert result.coords[1][1] == 0.4
    assert result.coords[1][2] == 0

    assert np.isnan(result.coords[5][0])
    assert np.isnan(result.coords[5][1])
    assert np.isnan(result.coords[5][2])

    assert len(result.tri_ids) == 6
    assert result.tri_ids[0] == 0
    assert result.tri_ids[1] == 1
    assert result.tri_ids[5] == -1

    assert len(result.distances) == 6
    assert result.distances[0] == 1

    hit_mask = result.hit_mask
    hit_coords = result.coords[hit_mask]
    assert len(hit_coords) == 4
    for h in hit_coords:
        assert np.isnan(h).any() == False


def test_multiple_ray_occlude():
    m = mesh_plane()
    m.build("medium")
    ray_origin, ray_direction = rays()
    result = m.occlusion(ray_origin, ray_direction)

    assert len(result) == 6
    assert result[0] == True
    assert result[1] == True
    assert result[2] == True
    assert result[3] == True
    assert result[4] == False
    assert result[5] == False
